@@ -1,19 +1,17 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 
 import sys
 import os
+from unittest.mock import Mock
 
-rtd = os.environ.get('READTHEDOCS', None) == 'True'
 cwd = os.getcwd()
 project_root = os.path.dirname(cwd)
 sys.path.insert(0, project_root)
 
 # -- Don't try to load CFFI (doesn't work on RTD) -----------------------------
 
-if rtd:
-    from mock import Mock
-    sys.modules['cffi'] = Mock()
-    sys.modules['cffi.verifier'] = Mock()
+sys.modules['cffi'] = Mock()
+sys.modules['cffi.verifier'] = Mock()
 import snimpy
 
 # -- General configuration ----------------------------------------------------
