@@ -1,8 +1,5 @@
 """interactive SNMP tool"""
 
-__author__ = 'Vincent Bernat'
-__email__ = 'bernat@luffy.cx'
-
 try:
     from snimpy._version import __version__  # nopep8
 except ImportError:
