@@ -30,8 +30,19 @@ import re
 import socket
 import inspect
 import threading
+import asyncio
 import ipaddress
-from pysnmp.entity.rfc3413.oneliner import cmdgen
+from pysnmp.hlapi.v3arch.asyncio import (
+    SnmpEngine, CommunityData, UsmUserData,
+    UdpTransportTarget, Udp6TransportTarget, ContextData,
+    ObjectType, ObjectIdentity,
+    get_cmd, set_cmd, walk_cmd, bulk_walk_cmd,
+    usmNoAuthProtocol, usmHMACMD5AuthProtocol, usmHMACSHAAuthProtocol,
+    usmHMAC128SHA224AuthProtocol, usmHMAC192SHA256AuthProtocol,
+    usmHMAC256SHA384AuthProtocol, usmHMAC384SHA512AuthProtocol,
+    usmNoPrivProtocol, usmDESPrivProtocol, usm3DESEDEPrivProtocol,
+    usmAesCfb128Protocol, usmAesCfb192Protocol, usmAesCfb256Protocol,
+)
 from pysnmp.proto import rfc1902, rfc1905
 from pysnmp.smi import error
 
@@ -79,6 +90,12 @@ class Session:
 
     _tls = threading.local()
 
+    def _run(self, coro):
+        """Run an async coroutine synchronously using a thread-local loop."""
+        if not hasattr(self._tls, "loop"):
+            self._tls.loop = asyncio.new_event_loop()
+        return self._tls.loop.run_until_complete(coro)
+
     def __init__(self, host,
                  community="public", version=2,
                  secname=None,
@@ -106,8 +123,8 @@ class Session:
         :param secname: Security name to use for SNMPv3 only.
         :type secname: str
         :param authprotocol: Authorization protocol to use for
-            SNMPv3. This can be `None` or either the string `SHA` or
-            `MD5`.
+            SNMPv3. This can be `None` or one of the strings `SHA`,
+            `MD5`, `SHA224`, `SHA256`, `SHA384` or `SHA512`.
         :type authprotocol: None or str
         :param authpassword: Authorization password if authorization
             protocol is not `None`.
@@ -132,51 +149,55 @@ class Session:
         self._version = version
         self._none = none
         if version == 3:
-            self._cmdgen = cmdgen.CommandGenerator()
+            self._engine = SnmpEngine()
             self._contextname = contextname
         else:
-            if not hasattr(self._tls, "cmdgen"):
-                self._tls.cmdgen = cmdgen.CommandGenerator()
-            self._cmdgen = self._tls.cmdgen
+            if not hasattr(self._tls, "engine"):
+                self._tls.engine = SnmpEngine()
+            self._engine = self._tls.engine
             self._contextname = None
         if version == 1 and none:
             raise ValueError("None-GET requests not compatible with SNMPv1")
 
         # Put authentication stuff in self._auth
         if version in [1, 2]:
-            self._auth = cmdgen.CommunityData(
+            self._auth = CommunityData(
                 community[0:30], community, version - 1)
         elif version == 3:
             if secname is None:
                 secname = community
             try:
                 authprotocol = {
-                    None: cmdgen.usmNoAuthProtocol,
-                    "MD5": cmdgen.usmHMACMD5AuthProtocol,
-                    "SHA": cmdgen.usmHMACSHAAuthProtocol,
-                    "SHA1": cmdgen.usmHMACSHAAuthProtocol
+                    None: usmNoAuthProtocol,
+                    "MD5": usmHMACMD5AuthProtocol,
+                    "SHA": usmHMACSHAAuthProtocol,
+                    "SHA1": usmHMACSHAAuthProtocol,
+                    "SHA224": usmHMAC128SHA224AuthProtocol,
+                    "SHA256": usmHMAC192SHA256AuthProtocol,
+                    "SHA384": usmHMAC256SHA384AuthProtocol,
+                    "SHA512": usmHMAC384SHA512AuthProtocol,
                 }[authprotocol]
             except KeyError:
                 raise ValueError("{} is not an acceptable authentication "
                                  "protocol".format(authprotocol))
             try:
                 privprotocol = {
-                    None: cmdgen.usmNoPrivProtocol,
-                    "DES": cmdgen.usmDESPrivProtocol,
-                    "3DES": cmdgen.usm3DESEDEPrivProtocol,
-                    "AES": cmdgen.usmAesCfb128Protocol,
-                    "AES128": cmdgen.usmAesCfb128Protocol,
-                    "AES192": cmdgen.usmAesCfb192Protocol,
-                    "AES256": cmdgen.usmAesCfb256Protocol,
+                    None: usmNoPrivProtocol,
+                    "DES": usmDESPrivProtocol,
+                    "3DES": usm3DESEDEPrivProtocol,
+                    "AES": usmAesCfb128Protocol,
+                    "AES128": usmAesCfb128Protocol,
+                    "AES192": usmAesCfb192Protocol,
+                    "AES256": usmAesCfb256Protocol,
                 }[privprotocol]
             except KeyError:
                 raise ValueError("{} is not an acceptable privacy "
                                  "protocol".format(privprotocol))
-            self._auth = cmdgen.UsmUserData(secname,
-                                            authpassword,
-                                            privpassword,
-                                            authprotocol,
-                                            privprotocol)
+            self._auth = UsmUserData(secname,
+                                     authpassword,
+                                     privpassword,
+                                     authprotocol,
+                                     privprotocol)
         else:
             raise ValueError("unsupported SNMP version {}".format(version))
 
@@ -192,11 +213,11 @@ class Session:
         else:
             port = 161
         if mo.group("ipv6"):
-            self._transport = cmdgen.Udp6TransportTarget((mo.group("ipv6"),
-                                                          port))
+            self._transport = self._run(
+                Udp6TransportTarget.create((mo.group("ipv6"), port)))
         elif mo.group("ipv4"):
-            self._transport = cmdgen.UdpTransportTarget((mo.group("ipv4"),
-                                                         port))
+            self._transport = self._run(
+                UdpTransportTarget.create((mo.group("ipv4"), port)))
         else:
             results = socket.getaddrinfo(mo.group("any"),
                                          port,
@@ -209,11 +230,18 @@ class Session:
             # an IPv4 address, use that. If not, use IPv6. If we want
             # to add an option to force IPv6, it is a good place.
             if [x for x in results if x[0] == socket.AF_INET]:
-                self._transport = cmdgen.UdpTransportTarget((mo.group("any"),
-                                                             port))
+                self._transport = self._run(
+                    UdpTransportTarget.create((mo.group("any"), port)))
             else:
-                self._transport = cmdgen.Udp6TransportTarget((mo.group("any"),
-                                                              port))
+                self._transport = self._run(
+                    Udp6TransportTarget.create((mo.group("any"), port)))
+
+        # Context data
+        if self._contextname:
+            self._contextdata = ContextData(
+                contextName=rfc1902.OctetString(self._contextname))
+        else:
+            self._contextdata = ContextData()
 
         # Bulk stuff
         self.bulk = bulk
@@ -227,6 +255,19 @@ class Session:
         if isinstance(value, rfc1905.EndOfMibView):
             raise SNMPEndOfMibView("End of MIB was reached")  # noqa: F821
 
+    def _check_error(self, errorIndication, errorStatus):
+        """Check for SNMP protocol errors in response"""
+        if errorIndication:
+            self._check_exception(errorIndication)
+            raise SNMPException(str(errorIndication))
+        if errorStatus:
+            exc = str(errorStatus.prettyPrint())
+            exc = re.sub(r'\W+', '', exc)
+            exc = "SNMP{}".format(exc[0].upper() + exc[1:])
+            if str(exc) in globals():
+                raise globals()[exc]
+            raise SNMPException(errorStatus.prettyPrint())
+
     def _convert(self, value):
         """Convert a PySNMP value to some native Python type"""
         try:
@@ -256,44 +297,52 @@ class Session:
         self._check_exception(value)
         raise NotImplementedError("unable to convert {}".format(repr(value)))
 
-    def _op(self, cmd, *oids):
-        """Apply an SNMP operation"""
-        kwargs = {}
-        if self._contextname:
-            kwargs['contextName'] = rfc1902.OctetString(self._contextname)
-        errorIndication, errorStatus, errorIndex, varBinds = cmd(
-            self._auth, self._transport, *oids, **kwargs)
-        if errorIndication:
-            self._check_exception(errorIndication)
-            raise SNMPException(str(errorIndication))
-        if errorStatus:
-            # We try to find a builtin exception with the same message
-            exc = str(errorStatus.prettyPrint())
-            exc = re.sub(r'\W+', '', exc)
-            exc = "SNMP{}".format(exc[0].upper() + exc[1:])
-            if str(exc) in globals():
-                raise globals()[exc]
-            raise SNMPException(errorStatus.prettyPrint())
-        if cmd in [self._cmdgen.getCmd, self._cmdgen.setCmd]:
-            results = [(tuple(name), val) for name, val in varBinds]
-        else:
-            results = [(tuple(name), val)
-                       for row in varBinds for name, val in row]
-            if len(results) > 0 and isinstance(results[-1][1],
-                                               rfc1905.EndOfMibView):
-                results = results[:-1]
-        if len(results) == 0:
-            if cmd not in [self._cmdgen.nextCmd, self._cmdgen.bulkCmd]:
-                raise SNMPException("empty answer")
-        return tuple([(oid, self._convert(val)) for oid, val in results])
-
     def get(self, *oids):
         """Retrieve an OID value using GET.
 
         :param oids: a list of OID to retrieve. An OID is a tuple.
         :return: a list of tuples with the retrieved OID and the raw value.
         """
-        return self._op(self._cmdgen.getCmd, *oids)
+        objecttypes = [ObjectType(ObjectIdentity(oid)) for oid in oids]
+        errorIndication, errorStatus, errorIndex, varBinds = self._run(
+            get_cmd(self._engine, self._auth, self._transport,
+                    self._contextdata, *objecttypes, lookupMib=False))
+        self._check_error(errorIndication, errorStatus)
+        results = [(tuple(name), self._convert(val))
+                   for name, val in varBinds]
+        if not results:
+            raise SNMPException("empty answer")
+        return tuple(results)
+
+    async def _walk_async(self, *oids):
+        """Collect results from GETNEXT-based walk."""
+        results = []
+        for oid in oids:
+            walker = walk_cmd(
+                self._engine, self._auth, self._transport,
+                self._contextdata,
+                ObjectType(ObjectIdentity(oid)), lookupMib=False)
+            async for result in walker:
+                errorIndication, errorStatus, errorIndex, varBinds = result
+                self._check_error(errorIndication, errorStatus)
+                for name, val in varBinds:
+                    results.append((tuple(name), val))
+        return results
+
+    async def _bulkwalk_async(self, bulk, *oids):
+        """Collect results from GETBULK-based walk."""
+        results = []
+        for oid in oids:
+            walker = bulk_walk_cmd(
+                self._engine, self._auth, self._transport,
+                self._contextdata, 0, bulk,
+                ObjectType(ObjectIdentity(oid)), lookupMib=False)
+            async for result in walker:
+                errorIndication, errorStatus, errorIndex, varBinds = result
+                self._check_error(errorIndication, errorStatus)
+                for name, val in varBinds:
+                    results.append((tuple(name), val))
+        return results
 
     def walkmore(self, *oids):
         """Retrieve OIDs values using GETBULK or GETNEXT. The method is called
@@ -306,20 +355,21 @@ class Session:
 
         """
         if self._version == 1 or not self.bulk:
-            return self._op(self._cmdgen.nextCmd, *oids)
-        args = [0, self.bulk] + list(oids)
-        try:
-            return self._op(self._cmdgen.bulkCmd, *args)
-        except SNMPTooBig:
-            # Let's try to ask for less values. We will never increase
-            # bulk again. We cannot increase it just after the walk
-            # because we may end up requesting everything twice (or
-            # more).
-            nbulk = self.bulk / 2 or False
-            if nbulk != self.bulk:
-                self.bulk = nbulk
-                return self.walk(*oids)
-            raise
+            results = self._run(self._walk_async(*oids))
+        else:
+            try:
+                results = self._run(self._bulkwalk_async(self.bulk, *oids))
+            except SNMPTooBig:
+                # Let's try to ask for less values. We will never increase
+                # bulk again. We cannot increase it just after the walk
+                # because we may end up requesting everything twice (or
+                # more).
+                nbulk = self.bulk / 2 or False
+                if nbulk != self.bulk:
+                    self.bulk = nbulk
+                    return self.walk(*oids)
+                raise
+        return tuple([(oid, self._convert(val)) for oid, val in results])
 
     def walk(self, *oids):
         """Walk from given OIDs but don't return any "extra" results. Only
@@ -346,8 +396,17 @@ class Session:
         """
         if len(args) % 2 != 0:
             raise ValueError("expect an even number of arguments for SET")
-        varbinds = zip(*[args[0::2], [v.pack() for v in args[1::2]]])
-        return self._op(self._cmdgen.setCmd, *varbinds)
+        objecttypes = [ObjectType(ObjectIdentity(oid), val.pack())
+                       for oid, val in zip(args[0::2], args[1::2])]
+        errorIndication, errorStatus, errorIndex, varBinds = self._run(
+            set_cmd(self._engine, self._auth, self._transport,
+                    self._contextdata, *objecttypes, lookupMib=False))
+        self._check_error(errorIndication, errorStatus)
+        results = [(tuple(name), self._convert(val))
+                   for name, val in varBinds]
+        if not results:
+            raise SNMPException("empty answer")
+        return tuple(results)
 
     def __repr__(self):
         return "{}(host={},version={})".format(
